###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging
from collections import defaultdict

from unitem.defaults import (TIGRFAM_TOP_HIT_SUFFIX,
                             PFAM_TOP_HIT_SUFFIX,
                             BINNING_METHOD_DIR)
from unitem.external.tophits import TopHitFile
from unitem.external.tigrfam_search import TigrfamSearch
from unitem.external.pfam_search import PfamSearch


class Markers():
    """Read marker information."""

    def __init__(self, marker_dir):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        pfam_bac_ms, pfam_ar_ms = PfamSearch(marker_dir, 1).get_marker_genes()
        tigr_bac_ms, tigr_ar_ms = TigrfamSearch(
            marker_dir, 1).get_marker_genes()

        self.bac_ms = pfam_bac_ms.union(tigr_bac_ms)
        self.ar_ms = pfam_ar_ms.union(tigr_ar_ms)

    @staticmethod
    def estimate_comp_cont(hmm_hits, markers):
        """Estimate genome completeness and contamination.

        completeness = (# identified genes)/(total marker genes)
        contamination = (multi-copy count)/(total marker genes)
        """

        present = 0
        multi_copy_count = 0
        for hmm_id, hits in hmm_hits.items():
            if hmm_id in markers:
                present += 1
                multi_copy_count += len(hits) - 1

        if len(markers) > 0:
            perc_comp = 100 * float(present) / len(markers)
            perc_cont = 100 * float(multi_copy_count) / len(markers)
        else:
            perc_comp = perc_cont = 0

        return perc_comp, perc_cont

    @staticmethod
    def read_top_hit_tables(top_hit_dir):
        """Read top hit tables for all bins in directory."""

        markers = defaultdict(lambda: defaultdict(list))

        for f in os.listdir(top_hit_dir):
            if not (f.endswith(PFAM_TOP_HIT_SUFFIX) or f.endswith(TIGRFAM_TOP_HIT_SUFFIX)):
                continue

            top_hit_file = os.path.join(top_hit_dir, f)
            top_hits = TopHitFile(top_hit_file)
            top_hits.read()

            bin_id = f.replace(PFAM_TOP_HIT_SUFFIX, '')
            bin_id = bin_id.replace(TIGRFAM_TOP_HIT_SUFFIX, '')

            for gene_id, cur_hit in top_hits.iter_hits():
                scaffold_id = gene_id.rpartition('_')[0]
                markers[bin_id][cur_hit.hmm_id].append(scaffold_id)

        return markers

    def evaluate(self, gene_table):
        """Evaluate completeness and contamination of genome using best domain-level marker sets."""

        bac_comp, bac_cont = Markers.estimate_comp_cont(
            gene_table,
            self.bac_ms
        )

        ar_comp, ar_cont = Markers.estimate_comp_cont(
            gene_table,
            self.ar_ms
        )

        if bac_comp + bac_cont > ar_comp + ar_cont:
            return 'Bacteria', bac_comp, bac_cont

        return 'Archaea', ar_comp, ar_cont

    def create_gene_table(self, contig_ids):
        """Create gene tables for a set of contigs."""

        gene_table = defaultdict(list)
        for cid in contig_ids:
            for mid in self.markers_on_contigs[cid]:
                gene_table[mid].append(cid)

        return gene_table

    def bin_quality(self, bin):
        """Estimate quality of bin."""

        gene_table = self.create_gene_table(bin)

        return self.evaluate(gene_table)

    def marker_gene_tables(self, profile_dir, binning_methods=None):
        """Get marker genes for bins across all binning methods."""

        binning_methods_dir = os.path.join(profile_dir, BINNING_METHOD_DIR)
        gene_tables = {}
        for bm in os.listdir(binning_methods_dir):
            if binning_methods and bm not in binning_methods:
                continue  # not in set of desired binning methods

            top_hit_dir = os.path.join(binning_methods_dir, bm)
            gene_tables[bm] = Markers.read_top_hit_tables(top_hit_dir)

        self._markers_on_contigs(gene_tables)

        return gene_tables

    def _markers_on_contigs(self, gene_tables):
        """Get markers on each contig."""

        self.markers_on_contigs = defaultdict(list)
        processed_contigs = set()
        for gene_tables in gene_tables.values():
            # process all bins identified for a given binning method
            scaffolds_in_binning_method = set()
            for bin_id in gene_tables:
                for marker_id, scaffold_ids in gene_tables[bin_id].items():
                    for scaffold_id in scaffold_ids:
                        if scaffold_id in processed_contigs:
                            continue

                        self.markers_on_contigs[scaffold_id].append(marker_id)

                    scaffolds_in_binning_method.update(scaffold_ids)

            processed_contigs.update(scaffolds_in_binning_method)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import logging

from unitem.common import run_cmd
from unitem.utils import check_on_path, make_sure_path_exists


class Bin():
    """Apply binning methods to an assembly."""

    def __init__(self, assembly_file, output_dir, min_contig_len, cpus):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        self.assembly_file = assembly_file
        self.output_dir = output_dir
        self.min_contig_len = min_contig_len
        self.cpus = cpus

        self.failed_methods = []

    def coverage(self, bam_files, cov_file):
        """Calculate coverage file for use by different binning methods."""

        self.bam_files = bam_files

        if cov_file:
            self.logger.info(f'Using coverage information in {cov_file}.')
            # check coverage file has correct format
            header = open(cov_file).readline().split('\t')
            if not ('contig' in header[0].lower()
                    and 'len' in header[1].lower()
                    and ('depth' in header[2].lower() or 'mean' in header[2].lower())):
                self.logger.error(
                    'Provided coverage file does not have the correct headers.')
                self.logger.error(
                    "Coverage file must have the format produced by 'jgi_summarize_bam_contig_depths'.")
                sys.exit(1)
            self.cov_file = cov_file

            if bam_files:
                self.logger.warning('BAM files are being ignored.')
        else:
            found = check_on_path(
                'jgi_summarize_bam_contig_depths', exit_on_fail=False)
            if not found:
                self.logger.error(
                    'jgi_summarize_bam_contig_depths is not on the system path.')
                self.logger.error('This script is provide with MetaBAT v2.')
                sys.exit(1)

            self.logger.info(
                f'Calculating coverage for {len(bam_files)} BAM files.')
            self.logger.info("Running jgi_summarize_bam_contig_depths script.")

            self.cov_file = os.path.join(self.output_dir, 'coverage.tsv')
            cmd = 'jgi_summarize_bam_contig_depths --minContigLength {} --minContigDepth 1 --outputDepth {} {}'.format(self.min_contig_len,
                                                                                                                       self.cov_file,
                                                                                                                       ' '.join(bam_files))

            run_cmd(cmd, program='jgi_summarize_bam_contig_depths')

    def check_on_path(self, options):
        """Check that all binning methods are on the system path."""

        if options.mb2:
            self.logger.info('Checking MetaBAT v2 dependencies.')
            check_on_path('metabat2')
        if options.gm2:
            self.logger.info('Checking GroopM v2 dependencies.')
            check_on_path('groopm2')
        if options.max40 or options.max107:
            self.logger.info('Checking MaxBin dependencies.')
            check_on_path('run_MaxBin.pl')
        if (options.mb_verysensitive
                or options.mb_sensitive
                or options.mb_specific
                or options.mb_veryspecific
                or options.mb_superspecific):
            self.logger.info('Checking MetaBAT dependencies.')
            check_on_path('metabat1')

    def run(self, options):
        """Run binning methods."""

        bin_file = os.path.join(self.output_dir, 'bin_dirs.tsv')
        bin_file_out = open(bin_file, 'w')

        if options.mb2:
            self.metabat2(bin_file_out)
        if options.gm2:
            self.groopm2(bin_file_out)
        if options.max40:
            self.maxbin(bin_file_out, 40)
        if options.max107:
            self.maxbin(bin_file_out, 107)
        if options.mb_verysensitive:
            self.metabat(bin_file_out, 'verysensitive')
        if options.mb_sensitive:
            self.metabat(bin_file_out, 'sensitive')
        if options.mb_specific:
            self.metabat(bin_file_out, 'specific')
        if options.mb_veryspecific:
            self.metabat(bin_file_out, 'veryspecific')
        if options.mb_superspecific:
            self.metabat(bin_file_out, 'superspecific')

        bin_file_out.close()

        self.logger.info(
            f'File with location of bin directories written to {bin_file}.')

        if self.failed_methods:
            self.logger.warning(
                f'The following methods failed to run: {" ".join(self.failed_methods)}')

    def _run_method(self, cmd, bin_dir, bin_file_out, binning_method):
        """Run binning method."""

        run_cmd(cmd, program=binning_method, exit_on_fail=False)

        bin_file_out.write(f'{binning_method}\t{os.path.abspath(bin_dir)}\n')

    def metabat2(self, bin_file_out):
        """Run MetaBAT v2."""

        self.logger.info("Running MetaBAT v2.")
        bin_dir = os.path.join(self.output_dir, 'metabat2')
        bin_prefix = os.path.join(bin_dir, 'mb2')
        cmd = 'metabat2 -t {} -m {} -i {} -a {} -o {}'.format(self.cpus,
                                                              self.min_contig_len,
                                                              self.assembly_file,
                                                              self.cov_file,
                                                              bin_prefix)

        self._run_method(cmd, bin_dir, bin_file_out, 'metabat2')

    def metabat(self, bin_file_out, preset):
        """Run MetaBAT."""

        self.logger.info(f"Running MetaBAT v1 with the {preset} preset.")
        bin_dir = os.path.join(self.output_dir, f'metabat_{preset}')
        bin_prefix = os.path.join(bin_dir, f'mb_{preset}')
        cmd = 'metabat1 -t {} -m {} -i {} -a {} -o {} --{}'.format(self.cpus,
                                                                   self.min_contig_len,
                                                                   self.assembly_file,
                                                                   self.cov_file,
                                                                   bin_prefix,
                                                                   preset)

        self._run_method(cmd, bin_dir, bin_file_out, f'metabat_{preset}')

    def groopm2(self, bin_file_out):
        """Run GroopM v2."""

        self.logger.info("Running GroopM v2 parse.")
        bin_dir = os.path.join(self.output_dir, 'groopm2')
        make_sure_path_exists(bin_dir)
        output_db = os.path.join(bin_dir, 'groopm.db')
        cmd = 'groopm2 parse -f -t {} -c {} --cov_file {} {} {}'.format(self.cpus,
                                                                        self.min_contig_len,
                                                                        self.cov_file,
                                                                        output_db,
                                                                        self.assembly_file)
        run_cmd(cmd, program='groopm2')

        self.logger.info("Running GroopM v2 core.")
        cmd = 'groopm2 core -f {} -c {} --save_dists'.format(output_db,
                                                             self.min_contig_len)
        run_cmd(cmd, program='groopm2')

        self.logger.info("Running GroopM v2 extract.")
        bin_prefix = os.path.join(bin_dir, 'gm2')
        cmd = 'groopm2 extract -p {} {} {}'.format(bin_prefix,
                                                   output_db,
                                                   self.assembly_file)

        self._run_method(cmd, bin_dir, bin_file_out, 'groopm2')

    def _create_maxbin_coverage_files(self, cov_file, output_dir):
        """Parse coverage information files required by MaxBin."""

        abund_list_file = os.path.join(output_dir, 'abund_files.lst')
        fout = open(abund_list_file, 'w')
        with open(cov_file) as f:
            headers = f.readline().rstrip().split('\t')
            bam_headers = [bh.replace(' ', '_').replace('/', '~')
                           for bh in headers[3::2]]

            fhs = []
            for bh in bam_headers:
                abund_file = os.path.abspath(
                    os.path.join(output_dir, bh + '.abund.tsv'))
                fhs.append(open(abund_file, 'w'))
                fout.write(abund_file + '\n')

            for line in f:
                line_split = line.rstrip().split('\t')
                contig_id = line_split[0]
                for fh_index, col_index in enumerate(range(3, len(line_split), 2)):
                    fhs[fh_index].write(
                        contig_id + '\t' + line_split[col_index] + '\n')

            for fh in fhs:
                fh.close()
        fout.close()

        return abund_list_file

    def maxbin(self, bin_file_out, num_markers):
        """Run MaxBin."""

        bin_dir = os.path.join(self.output_dir, f'maxbin_ms{num_markers}')
        make_sure_path_exists(bin_dir)
        cov_file_dir = os.path.join(bin_dir, 'coverage_files')
        make_sure_path_exists(cov_file_dir)
        abund_list_file = self._create_maxbin_coverage_files(
            self.cov_file, cov_file_dir)

        self.logger.info(f"Running MaxBin v2 with {num_markers} markers.")
        bin_prefix = os.path.join(bin_dir, f'max{num_markers}')
        cmd = f'run_MaxBin.pl -min_contig_length {self.min_contig_len}'
        cmd += ' -thread {} -markerset {} -contig {} -out {} -abund_list {}'.format(self.cpus,
                                                                                    num_markers,
                                                                                    self.assembly_file,
                                                                                    bin_prefix,
                                                                                    abund_list_file)

        self._run_method(cmd, bin_dir, bin_file_out, f'maxbin_ms{num_markers}')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging

import svgwrite


class TreeCommonBases():
    """Create dendrogram showing common bases between bins."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        # size of labels
        self.font_size = 8
        self.scale_font_size = 6

        # width and height of each gene cell
        self.row_height = 1.3*self.font_size
        self.scale_interval_width = 15

        self.label_offset = 2

    def _render_scale(self, dwg):
        """Render scale."""

        scale_group = svgwrite.container.Group(id='scale')
        dwg.add(scale_group)

        scale_text_group = svgwrite.container.Group(id='scale_text')
        dwg.add(scale_text_group)

        # draw baseline
        scale_y_pos = self.fig_size_y + 0.5*self.row_height
        scale = dwg.path("M%f,%f" % (0, scale_y_pos + 0.25*self.row_height))
        scale.push("L%f,%f" % (0, scale_y_pos))
        scale.push("L%f,%f" % (self.fig_size_x, scale_y_pos))
        scale.push("L%f,%f" %
                   (self.fig_size_x, scale_y_pos + 0.25*self.row_height))
        scale.fill(color='none')
        scale.stroke(color='black', width=0.5)
        scale_group.add(scale)

        for s in [50, 60, 70, 80, 90, 100]:
            xpos = self.fig_size_x - ((100-s)/10)*self.scale_interval_width
            t = dwg.text(int(s),
                         x=[(xpos)],
                         y=[(scale_y_pos + 0.75*self.row_height + self.label_offset)],
                         font_size="%fpt" % self.scale_font_size,
                         text_anchor='middle',
                         fill='rgb(0,0,0)')
            scale_text_group.add(t)

            if s != 50 and s != 100:
                tick = dwg.line(start=(xpos, scale_y_pos),
                                end=(xpos, scale_y_pos + 0.25*self.row_height),
                                fill='black',
                                stroke_width=0.5)
                tick.stroke(color='black')
                scale_group.add(tick)

    def _render_labels(self, dwg, bin_id, common_bases):
        """Render labels."""

        label_group = svgwrite.container.Group(id='labels')
        dwg.add(label_group)

        t = dwg.text(bin_id,
                     x=[(self.fig_size_x + self.label_offset)],
                     y=[(self.fig_size_y + 0.5*self.font_size)],
                     font_size="%fpt" % self.font_size,
                     direction='ltr',
                     fill='rgb(0,0,0)')
        label_group.add(t)

        y = self.fig_size_y
        for bm, bid, cb in common_bases:
            y -= self.row_height
            t = dwg.text(bm,
                         x=[(self.fig_size_x + self.label_offset)],
                         y=[(y + 0.5*self.font_size)],
                         font_size="%fpt" % self.font_size,
                         direction='ltr',
                         fill='rgb(0,0,0)')
            label_group.add(t)

    def _render_tree(self, dwg, common_bases):
        """Render tree."""

        tree_group = svgwrite.container.Group(id='tree')
        dwg.add(tree_group)

        x_start = self.fig_size_x
        y_start = self.fig_size_y
        width_start = 0
        for r, (_bm, _bid, cb) in enumerate(common_bases):
            width = ((100-cb)/10)*self.scale_interval_width
            delta_width = width - width_start

            branch = dwg.path("M%f,%f" % (x_start, y_start))
            branch.push("L%f,%f" % (x_start-delta_width, y_start))
            branch.push("L%f,%f" % (x_start-delta_width,
                        self.fig_size_y - (r+1)*self.row_height))
            branch.push("L%f,%f" % (self.fig_size_x,
                        self.fig_size_y - (r+1)*self.row_height))
            branch.fill(color='none')
            branch.stroke(color='black', width=1)
            tree_group.add(branch)

            x_start = self.fig_size_x - width
            y_start = self.fig_size_y - r*self.row_height - 0.5*self.row_height
            width_start = width

        # render root branch
        root = dwg.line(start=(x_start, y_start),
                        end=(0, y_start),
                        fill='black',
                        stroke_width=1)
        root.stroke(color='black')
        tree_group.add(root)

    def plot(self, bin_id, common_bases, output_plot):
        """Create plot.

        Parameters
        ----------
        bin_id : str
          Identifier of bin compared to other bins.
        common_bases : d[binning method][bin id] -> percent common bases
          Percentage of common bases for each binning method.
        output_plot : str
          Desired output file.
        """

        # setup SVG drawing
        start_y = 0

        if not output_plot.endswith('.svg'):
            output_plot += '.svg'

        self.fig_size_x = 5*self.scale_interval_width  # 50, 60, 70, 80, 90, 100
        self.fig_size_y = start_y + len(common_bases)*self.row_height

        dwg = svgwrite.Drawing(filename=output_plot,
                               size=(self.fig_size_x, self.fig_size_y),
                               profile='full')
        dwg.set_desc(title='UniteM shared base pair tree.')

        self._render_scale(dwg)
        self._render_labels(dwg, bin_id, common_bases)
        self._render_tree(dwg, common_bases)

        dwg.save()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import shutil
import subprocess
import tempfile
import multiprocessing as mp

from unitem.seq_io import read_fasta, write_fasta
from unitem.utils import make_sure_path_exists

from numpy import (sum as np_sum, zeros as np_zeros, bool as np_bool)


AA_GENE_FILE_SUFFIX = '_genes.faa'
NT_GENE_FILE_SUFFIX = '_genes.fna'


class ProdigalException(Exception):
    """Exception for errors associated with running Prodigal."""

    def __init__(self, message=''):
        Exception.__init__(self, message)


class Prodigal(object):
    """Perform gene prediction using Prodigal."""

    def __init__(self, cpus):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')

        self.cpus = cpus

    @staticmethod
    def version():
        try:
            proc = subprocess.Popen(['prodigal', '-v'],
                                    stderr=subprocess.PIPE,
                                    encoding='utf-8')

            _output, error = proc.communicate()

            return error.split('\n')[1].split()[1][1:-1]
        except:
            return "(version unavailable)"

    def _run_prodigal(self, gid, genome_file):
        """Run Prodigal."""

        # check if genome has previously been processed
        trans_table_file = os.path.join(
            self.output_dir,
            f'{gid}_translation_table.tsv')
        aa_gene_file = os.path.join(
            self.output_dir, f'{gid}{AA_GENE_FILE_SUFFIX}')
        nt_gene_file = os.path.join(
            self.output_dir, f'{gid}{NT_GENE_FILE_SUFFIX}')
        gff_file = os.path.join(self.output_dir, f'{gid}.gff')

        if (os.path.exists(trans_table_file)
            and os.path.exists(aa_gene_file)
            and os.path.exists(nt_gene_file)
                and os.path.exists(gff_file)):
            # use previously calculated results
            return aa_gene_file, nt_gene_file, gff_file, trans_table_file

        seqs = read_fasta(genome_file)

        if len(seqs) == 0:
            self.logger.warning(
                f'Cannot call Prodigal on empty genome: {genome_file}')
            return None

        with tempfile.TemporaryDirectory() as tmp_dir:

            # determine number of bases
            total_bases = sum(len(seq) for seq in seqs.values())

            # Prodigal requires an uncompressed FASTA file
            prodigal_input = genome_file
            if genome_file.endswith('.gz'):
                prodigal_input = os.path.join(tmp_dir, f'{gid}.fna')
                write_fasta(seqs, prodigal_input)

            # there may be ^M character in the input file,
            # the following code is similar to dos2unix command to remove
            # those special characters.
            with open(prodigal_input, 'r') as fh:
                text = fh.read().replace('\r\n', '\n')

            processed_prodigal_input = os.path.join(
                tmp_dir,
                os.path.basename(prodigal_input))
            with open(processed_prodigal_input, 'w') as fh:
                fh.write(text)

            # call genes under different translation tables
            table_coding_density = {4: -1, 11: -1}
            for trans_table in table_coding_density:
                tt_tmp_dir = os.path.join(tmp_dir, str(trans_table))
                os.makedirs(tt_tmp_dir)
                aa_gene_file_tmp = os.path.join(
                    tt_tmp_dir,
                    f'{gid}{AA_GENE_FILE_SUFFIX}')
                nt_gene_file_tmp = os.path.join(
                    tt_tmp_dir,
                    f'{gid}{NT_GENE_FILE_SUFFIX}')
                gff_file_tmp = os.path.join(tt_tmp_dir, gid + '.gff')

                # check if there is sufficient bases to calculate prodigal parameters
                cmd = ['prodigal', '-m', '-q', '-f', 'gff']
                if total_bases < 100000:
                    cmd += ['-p', 'meta']  # use best pre-calculated parameters
                else:
                    cmd += ['-p', 'single']  # estimate parameters from data

                cmd += ['-g', str(trans_table)]
                cmd += ['-a', str(aa_gene_file_tmp)]
                cmd += ['-d', str(nt_gene_file_tmp)]
                cmd += ['-i', str(processed_prodigal_input)]

                with open(gff_file_tmp, 'wb') as stdout:
                    proc = subprocess.Popen(cmd,
                                            stdout=stdout,
                                            encoding='utf-8')

                    proc.communicate()
                    exit_code = proc.wait()
                    if exit_code != 0:
                        self.logger.error(
                            'Prodigal returned non-zero exit code.')
                        self.logger.error(f'Skipping genome {gid}.')
                        return None

                prodigalParser = ProdigalGeneFeatureParser(gff_file_tmp)

                codingBases = 0
                for seq_id, _seq in seqs.items():
                    codingBases += prodigalParser.coding_bases(seq_id)

                codingDensity = float(codingBases) / total_bases
                table_coding_density[trans_table] = codingDensity

            # determine best translation table
            trans_table_out = open(trans_table_file, 'w')
            trans_table_out.write('Table\tCoding density\n')
            trans_table_out.write(f'4\t{table_coding_density[4]:.3f}\n')
            trans_table_out.write(f'11\t{table_coding_density[11]:.3f}\n')

            best_tt = 11
            if (table_coding_density[4] - table_coding_density[11] > 0.05) and table_coding_density[4] > 0.7:
                best_tt = 4

            trans_table_out.write(f'\nSelected table\t{best_tt}\n')

            # copy results for best translation table to desired output files
            best_tt_dir = os.path.join(tmp_dir, str(best_tt))
            shutil.copyfile(os.path.join(
                best_tt_dir, f'{gid}{AA_GENE_FILE_SUFFIX}'), aa_gene_file)
            shutil.copyfile(os.path.join(
                best_tt_dir, f'{gid}{NT_GENE_FILE_SUFFIX}'), nt_gene_file)
            shutil.copyfile(os.path.join(best_tt_dir, f'{gid}.gff'), gff_file)

        return aa_gene_file, nt_gene_file, gff_file, trans_table_file

    def _worker(self, out_dict, worker_queue, writer_queue):
        """Run Prodigal in parallel across genomes."""

        while True:
            data = worker_queue.get(block=True, timeout=None)
            if data is None:
                break

            gid, genome_file = data

            out_files = self._run_prodigal(gid, genome_file)

            # Only proceed if an error didn't occur when running Prodigal
            if out_files:
                prodigal_output = {"aa_gene_path": out_files[0],
                                   "nt_gene_path": out_files[1],
                                   "gff_path": out_files[2],
                                   "translation_table_path": out_files[3]}

                out_dict[gid] = prodigal_output

            writer_queue.put(gid)

    def _writer(self, num_genomes, writer_queue):
        """Track progress."""

        num_processed = 0
        while True:
            data = writer_queue.get(block=True, timeout=None)
            if data is None:
                break

            num_processed += 1
            status = f' -> processed {num_processed:,} of {num_genomes:,} genomes'
            sys.stdout.write(f'{status}\r')
            sys.stdout.flush()

    def run(self, genomic_files, output_dir):
        """Run Prodigal across a set of genomes."""

        self.output_dir = output_dir
        make_sure_path_exists(output_dir)

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for gid, genome_file in genomic_files.items():
            worker_queue.put((gid, genome_file))

        for _ in range(self.cpus):
            worker_queue.put(None)

        worker_proc = []
        writer_proc = None
        try:
            manager = mp.Manager()
            out_dict = manager.dict()

            worker_proc = [mp.Process(target=self._worker, args=(out_dict,
                                                                 worker_queue,
                                                                 writer_queue))
                           for _ in range(self.cpus)]
            writer_proc = mp.Process(target=self._writer, args=(len(genomic_files),
                                                                writer_queue))

            writer_proc.start()
            for p in worker_proc:
                p.start()

            for p in worker_proc:
                p.join()

                # Gracefully terminate the program.
                if p.exitcode != 0:
                    raise ProdigalException(
                        'Prodigal returned non-zero exit code.')

            writer_queue.put(None)
            writer_proc.join()
        except Exception as e:
            for p in worker_proc:
                p.terminate()

            if writer_proc:
                writer_proc.terminate()

            raise ProdigalException(
                f'Exception caught while running Prodigal: {e}')

        # report genomes which failed to have any genes called
        result_dict = dict()
        failed_gids = list()
        for gid, gid_dict in out_dict.items():
            if os.path.getsize(gid_dict['aa_gene_path']) <= 1:
                failed_gids.append(gid)
            else:
                result_dict[gid] = gid_dict

        if len(failed_gids) > 0:
            self.logger.warning(f'Skipping {len(failed_gids)} of {len(genomic_files)} '
                                'genomes as no genes were called by Prodigal.')

        return result_dict


class ProdigalGeneFeatureParser(object):
    """Parses prodigal gene feature files (GFF) output."""

    def __init__(self, filename):
        """Initialization.

        Parameters
        ----------
        filename : str
            GFF file to parse.
        """

        self.genes = {}
        self.last_coding_base = {}

        self._parseGFF(filename)

        self.coding_base_masks = {}
        for seq_id in self.genes:
            self.coding_base_masks[seq_id] = self._build_coding_base_mask(
                seq_id)

    def _parseGFF(self, filename):
        """Parse genes from GFF file.

        Parameters
        ----------
        filename : str
            GFF file to parse.
        """
        bGetTranslationTable = True
        with open(filename, 'r') as fh:
            for line in fh:
                if bGetTranslationTable and line.startswith('# Model Data'):
                    data_model_info = line.split(':')[1].strip().split(';')
                    dict_data_model = {}
                    for item in data_model_info:
                        k = item.split('=')[0]
                        v = item.split('=')[1]
                        dict_data_model[k] = v

                    self.translationTable = int(
                        dict_data_model.get('transl_table'))
                    bGetTranslationTable = False

                if line[0] == '#':
                    continue

                line_split = line.split('\t')
                seq_id = line_split[0]
                if seq_id not in self.genes:
                    geneCounter = 0
                    self.genes[seq_id] = {}
                    self.last_coding_base[seq_id] = 0

                geneId = seq_id + '_' + str(geneCounter)
                geneCounter += 1
                start = int(line_split[3])
                end = int(line_split[4])

                self.genes[seq_id][geneId] = [start, end]
                self.last_coding_base[seq_id] = max(
                    self.last_coding_base[seq_id], end)

    def _build_coding_base_mask(self, seq_id):
        """Build mask indicating which bases in a sequences are coding.

        Parameters
        ----------
        seq_id : str
            Unique id of sequence.
        """

        # safe way to calculate coding bases as it accounts
        # for the potential of overlapping genes
        coding_base_mask = np_zeros(
            self.last_coding_base[seq_id], dtype=np_bool)
        for pos in self.genes[seq_id].values():
            coding_base_mask[pos[0]:pos[1] + 1] = True

        return coding_base_mask

    def coding_bases(self, seq_id, start=0, end=None):
        """Calculate number of coding bases in sequence between [start, end).

        To process the entire sequence set start to 0, and
        end to None.

        Parameters
        ----------
        seq_id : str
            Unique id of sequence.
        start : int
            Start calculation at this position in sequence.
        end : int
            End calculation just before this position in the sequence.
        """

        # check if sequence has any genes
        if seq_id not in self.genes:
            return 0

        # set end to last coding base if not specified
        if end is None:
            end = self.last_coding_base[seq_id]

        return np_sum(self.coding_base_masks[seq_id][start:end])



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import math
import logging

from unitem.utils import alphanumeric_sort

import svgwrite


class PlotCommonBases():
    """Create heatmap showing percentage of common bases between bins."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        # size of table labels
        self.font_size = 8

        # size of labels in gene cells
        self.cell_font_size = 6

        # width and height of each gene cell
        self.row_height = 2.5*self.font_size
        self.col_width = self.row_height
        self.cell_offset = 2  # border around each gene cell

        # expanded width of columns with comp and cont information
        self.qual_col_width = 4*self.font_size

    def _render_label_col(self, dwg,
                          labels,
                          start_x,
                          start_y,
                          group_id):
        """Render column of labels."""

        label_group = svgwrite.container.Group(id=group_id)
        dwg.add(label_group)

        for i, label in enumerate(labels):
            t = dwg.text(label,
                         x=[(start_x)],
                         y=[(start_y + i*self.row_height)],
                         font_size="%fpt" % self.font_size,
                         text_anchor='end',
                         direction='ltr',
                         fill='black')

            label_group.add(t)

    def _render_label_row(self, dwg,
                          labels,
                          label_start_x,
                          row_start_y,
                          group_id,
                          rotation=-45):
        """Render column of labels."""

        label_group = svgwrite.container.Group(id=group_id)
        dwg.add(label_group)

        for i, label in enumerate(labels):
            x = label_start_x + (i+0.5)*self.col_width
            t = dwg.text(label,
                         x=[(x)],
                         y=[(row_start_y)],
                         font_size="%fpt" % self.font_size,
                         text_anchor='start',
                         direction='ltr',
                         fill='black')

            t.rotate(rotation, (x, row_start_y))

            label_group.add(t)

    def _render_genome_quality_cols(self,
                                    dwg,
                                    bin_labels,
                                    quality,
                                    header_start_y,
                                    label_start_x,
                                    label_start_y):
        """Plot genome completeness and contamination columns."""

        # render header
        quality_group = svgwrite.container.Group(id='genome_quality')
        dwg.add(quality_group)

        for c, gene_name in enumerate(['Completeness (%)', 'Contamination (%)']):
            x = label_start_x + (c + 0.5)*self.qual_col_width
            t = dwg.text(gene_name,
                         x=[(x)],
                         y=[(header_start_y)],
                         font_size="%fpt" % self.font_size,
                         text_anchor='start',
                         direction='ltr',
                         fill='black')

            t.rotate(-45, (x, header_start_y))

            quality_group.add(t)

        # render rows
        for r, bin_label in enumerate(bin_labels):
            for c, q in enumerate(quality[bin_label]):
                t = dwg.text('%.1f' % q,
                             x=[(label_start_x + (c+0.5)*self.qual_col_width)],
                             y=[(label_start_y + r*self.row_height)],
                             font_size="%fpt" % self.font_size,
                             text_anchor='middle',
                             direction='ltr',
                             fill='black')
                quality_group.add(t)

    def _cell_properties(self, perc_common_bases):
        """Get desired color and size of heat map cell for a given percentage of common bases."""

        color = 'rgb(255,255,255)'
        size = self.col_width-2*self.cell_offset

        if perc_common_bases >= 100:
            color = 'rgb(165,15,21)'
        elif perc_common_bases >= 90:
            color = 'rgb(222,45,38)'
            size = math.sqrt(0.9)*size
        elif perc_common_bases >= 80:
            color = 'rgb(251,106,74)'
            size = math.sqrt(0.8)*size
        elif perc_common_bases >= 70:
            color = 'rgb(252,146,114)'
            size = math.sqrt(0.7)*size
        elif perc_common_bases >= 60:
            color = 'rgb(252,187,161)'
            size = math.sqrt(0.6)*size
        elif perc_common_bases >= 50:
            color = 'rgb(254,229,217)'
            size = math.sqrt(0.5)*size

        return color, size

    def _render_row(self,
                    dwg,
                    bin_labels,
                    bm_labels,
                    common_bases,
                    start_x,
                    start_y):
        """Render rows showing percentage of common bases."""

        table_group = svgwrite.container.Group(id='table')
        dwg.add(table_group)

        for r, bin_label in enumerate(bin_labels):
            for c, bm_label in enumerate(bm_labels):
                perc_cb = common_bases[bin_label].get(bm_label, 0)
                color, size = self._cell_properties(perc_cb)

                # plot heat map box
                x = start_x + c*self.col_width
                y = start_y + r*self.row_height

                if perc_cb > 0:
                    base_color, base_size = self._cell_properties(0)
                    rect = dwg.rect(insert=(x+0.5*(base_size - size),
                                            y+0.5*(base_size - size)),
                                    size=(size, size),
                                    fill=color)
                    rect.stroke(color='rgb(196,196,196)', width=0.1)
                    table_group.add(rect)

    def _render_legend(self, dwg):
        """Render legend."""

        legend_group = svgwrite.container.Group(id='legend')
        dwg.add(legend_group)

        x = self.fig_size_x
        y = 0
        _base_color, base_size = self._cell_properties(0)
        for perc_common_bases in [100, 90, 80, 70, 60, 50]:
            color, size = self._cell_properties(perc_common_bases)
            rect = dwg.rect(insert=(x+0.5*(base_size - size),
                                    y+0.5*(base_size - size)),
                            size=(size, size),
                            fill=color)
            rect.stroke(color='rgb(196,196,196)', width=0.1)
            legend_group.add(rect)

            legend_str = '%d' % perc_common_bases
            if perc_common_bases != 100:
                legend_str = '>' + legend_str

            t = dwg.text(legend_str,
                         x=[(x + base_size + self.cell_offset)],
                         y=[(y + 0.5*base_size + 0.5*self.font_size)],
                         font_size="%fpt" % self.font_size,
                         direction='ltr',
                         fill='rgb(0,0,0)')
            legend_group.add(t)

            y += size + self.cell_offset

    def plot(self, common_bases, quality, output_plot):
        """Create plot.

        Parameters
        ----------
        common_bases : d[unitem bid][binning method] -> percent common bases
          Percentage of common bases for each binning method.
        quality : d[unitem id] -> (completeness, contamination)
          Completeness and contamination of bins.
        output_plot : str
          Desired output file.
        """

        # sort bin labels
        bin_labels = alphanumeric_sort(common_bases)

        # sort binning method labels
        binning_methods = set()
        for bid in common_bases:
            for bm in common_bases[bid]:
                binning_methods.add(bm)
        bm_labels = alphanumeric_sort(binning_methods)

        # setup SVG drawing
        table_start_x = 0
        table_start_y = 0

        if not output_plot.endswith('.svg'):
            output_plot += '.svg'

        self.fig_size_x = table_start_x + len(bm_labels)*self.col_width
        self.fig_size_x += 2*self.qual_col_width + 0.5*self.col_width
        self.fig_size_y = table_start_y + len(bin_labels)*self.row_height

        dwg = svgwrite.Drawing(filename=output_plot,
                               size=(self.fig_size_x, self.fig_size_y),
                               profile='full',
                               style='font-family:Arial')
        dwg.set_desc(title='UniteM shared base pair plot.')

        # render legend
        self._render_legend(dwg)

        # render binning method labels
        label_start_x = table_start_x
        label_start_y = table_start_y + 0.5*self.row_height + 0.45*self.font_size
        self._render_label_col(dwg,
                               bin_labels,
                               label_start_x,
                               label_start_y,
                               'bin_labels')

        # render completeness and contamination
        header_start_y = table_start_y - 0.5*self.font_size
        label_start_x = table_start_x + 0.5*self.col_width
        label_start_y = table_start_y + 0.5*self.row_height + 0.45*self.font_size
        self._render_genome_quality_cols(dwg,
                                         bin_labels,
                                         quality,
                                         header_start_y,
                                         label_start_x,
                                         label_start_y)

        # move start of table to account for genome quality info
        table_start_x += 2*self.qual_col_width + 0.5*self.col_width

        # write header line
        label_start_x = table_start_x
        header_row_start_y = table_start_y - 0.5*self.font_size
        self._render_label_row(dwg,
                               bm_labels,
                               label_start_x,
                               header_row_start_y,
                               'binning_method_labels')

        # render gene complement row
        self._render_row(dwg,
                         bin_labels,
                         bm_labels,
                         common_bases,
                         table_start_x,
                         table_start_y)

        dwg.save()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import inspect
import multiprocessing as mp

from unitem.external.tophits import TopHitPfamFile
from unitem.external.prodigal import AA_GENE_FILE_SUFFIX
from unitem.external.pypfam.Scan.PfamScan import PfamScan

PFAM_SUFFIX = "_pfam.tsv"
PFAM_OUT = '_pfam.out'


class PfamException(Exception):
    """Exception for errors associated with running HMMER with Pfam marker genes."""

    def __init__(self, message=''):
        Exception.__init__(self, message)


class PfamSearch(object):
    """Runs pfam_search.pl over a set of genomes."""

    def __init__(self, marker_dir, cpus):
        """Initialization."""

        self.cpus = cpus

        self.logger = logging.getLogger('timestamp')

        if marker_dir is None:
            self.marker_dir = PfamSearch.hmm_dir()
        else:
            self.marker_dir = marker_dir

        self.logger.info(f' - using Pfam HMMs in: {self.marker_dir}')

    @staticmethod
    def hmm_dir():
        """Get directory with PFAM HMM model files."""

        src_file_path = os.path.split(inspect.getfile(lambda: None))[0]
        hmm_dir = os.path.abspath(
            os.path.join(src_file_path, '..', '..', 'markers'))

        return hmm_dir

    def get_marker_genes(self):
        """Get bacterial and archaeal PFAM marker genes."""

        bac_ms = set()
        with open(os.path.join(self.marker_dir, 'pfam_bac.lst')) as f:
            for line in f:
                bac_ms.add(line.strip())

        ar_ms = set()
        with open(os.path.join(self.marker_dir, 'pfam_ar.lst')) as f:
            for line in f:
                ar_ms.add(line.strip())

        return bac_ms, ar_ms

    def _top_hit(self, pfam_file, output_dir):
        """Determine top hits to PFAMs.

        A gene may be assigned to multiple
        PFAM families from the same clan. The
        search_pfam.pl script takes care of
        most of these issues and here the results
        are simply parsed.

        Parameters
        ----------
        pfam_file : str
            Name of file containing hits to PFAM HMMs.
        """

        filename = os.path.basename(pfam_file)
        gid = filename.replace(PFAM_SUFFIX, '')
        tophit_file = TopHitPfamFile(output_dir, gid)

        with open(pfam_file, 'r') as fh_pfam:
            for line in fh_pfam:
                if line[0] == '#' or not line.strip():
                    continue

                line_split = line.split()
                gene_id = line_split[0]
                hmm_id = line_split[5]
                evalue = float(line_split[12])
                bitscore = float(line_split[11])
                tophit_file.add_hit(gene_id, hmm_id, evalue, bitscore)

        tophit_file.write()

    def _worker(self, gene_dir, output_dir, queue_in, queue_out):
        """Process each genome in parallel."""

        while True:
            gene_file = queue_in.get(block=True, timeout=None)
            if gene_file is None:
                break

            filename = os.path.basename(gene_file)
            output_hit_file = os.path.join(gene_dir, filename.replace(AA_GENE_FILE_SUFFIX,
                                                                      PFAM_SUFFIX))

            if os.path.exists(output_hit_file):
                # use previously calculated results
                queue_out.put(gene_file)
                continue

            # identify Pfam marker genes
            pfam_scan = PfamScan(cpu=self.cpus_per_genome,
                                 fasta=gene_file,
                                 dir=self.marker_dir)
            pfam_scan.search()
            pfam_scan.write_results(output_hit_file, None, None, None, None)

            # identify top hit for each gene
            self._top_hit(output_hit_file, output_dir)

            # allow results to be processed or written to file
            queue_out.put(gene_file)

    def _writer(self, num_genomes, writer_queue):
        """Track progress."""

        num_processed = 0
        while True:
            data = writer_queue.get(block=True, timeout=None)
            if data is None:
                break

            num_processed += 1
            status = f' -> processed {num_processed:,} of {num_genomes:,} genomes'
            sys.stdout.write(f'{status}\r')
            sys.stdout.flush()

    def run(self, gene_files, gene_dir, output_dir):
        """Annotate genes with Pfam HMMs.

        Parameters
        ----------
        gene_files : iterable
            Gene files in FASTA format to process.
        """

        if len(gene_files) == 0:
            self.logger.error('No genomes to process.')
            sys.exit(1)

        self.cpus_per_genome = max(1, int(self.cpus / len(gene_files)))

        # populate worker queue with data to process
        worker_queue = mp.Queue()
        writer_queue = mp.Queue()

        for f in gene_files:
            worker_queue.put(f)

        for _ in range(self.cpus):
            worker_queue.put(None)

        try:
            worker_proc = [mp.Process(target=self._worker, args=(
                gene_dir,
                output_dir,
                worker_queue,
                writer_queue)) for _ in range(self.cpus)]
            write_proc = mp.Process(target=self._writer, args=(len(gene_files),
                                                               writer_queue))

            write_proc.start()

            for p in worker_proc:
                p.start()

            for p in worker_proc:
                p.join()

                if p.exitcode != 0:
                    raise PfamException(
                        'HMMER returned non-zero exit code.')

            writer_queue.put(None)
            write_proc.join()
        except Exception:
            for p in worker_proc:
                p.terminate()

            write_proc.terminate()
            raise



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import re
import logging
import ntpath
import argparse
import errno



class CustomHelpFormatter(argparse.HelpFormatter):
    """Provide a customized format for help output.

    http://stackoverflow.com/questions/9642692/argparse-help-without-duplicate-allcaps
    """

    def _get_help_string(self, action):
        """Place default value in help string."""
        h = action.help
        if '%(default)' not in action.help:
            if action.default != '' and action.default != [] and action.default is not None and type(action.default) != bool:
                if action.default is not argparse.SUPPRESS:
                    defaulting_nargs = [
                        argparse.OPTIONAL, argparse.ZERO_OR_MORE]

                    if action.option_strings or action.nargs in defaulting_nargs:
                        if '\n' in h:
                            lines = h.splitlines()
                            lines[0] += ' (default: %(default)s)'
                            h = '\n'.join(lines)
                        else:
                            h += ' (default: %(default)s)'
            return h

    def _format_action_invocation(self, action):
        """Removes duplicate ALLCAPS with positional arguments."""
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar

        else:
            parts = []

            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    parts.append(option_string)

                return '%s %s' % (', '.join(parts), args_string)

            return ', '.join(parts)

    def _get_default_metavar_for_optional(self, action):
        return action.dest.upper()

    def _get_default_metavar_for_positional(self, action):
        return action.dest


def logger_setup(log_dir, log_file, program_name, version, silent):
    """Setup loggers.

    Two logger are setup which both print to the stdout and a 
    log file when the log_dir is not None. The first logger is
    named 'timestamp' and provides a timestamp with each call,
    while the other is named 'no_timestamp' and does not prepend
    any information. The attribution 'is_silent' is also added
    to each logger to indicate if the silent flag is thrown.

    Parameters
    ----------
    log_dir : str
        Output directory for log file.
    log_file : str
        Desired name of log file.
    program_name : str
        Name of program.
    version : str
        Program version number.
    silent : boolean
        Flag indicating if output to stdout should be suppressed.
    """

    # setup general properties of loggers
    timestamp_logger = logging.getLogger('timestamp')
    timestamp_logger.setLevel(logging.DEBUG)
    log_format = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                   datefmt="%Y-%m-%d %H:%M:%S")

    no_timestamp_logger = logging.getLogger('no_timestamp')
    no_timestamp_logger.setLevel(logging.DEBUG)

    # setup logging to console
    timestamp_stream_logger = logging.StreamHandler(sys.stdout)
    timestamp_stream_logger.setFormatter(log_format)
    timestamp_logger.addHandler(timestamp_stream_logger)

    no_timestamp_stream_logger = logging.StreamHandler(sys.stdout)
    no_timestamp_stream_logger.setFormatter(None)
    no_timestamp_logger.addHandler(no_timestamp_stream_logger)

    timestamp_logger.is_silent = False
    no_timestamp_stream_logger.is_silent = False
    if silent:
        timestamp_logger.is_silent = True
        timestamp_stream_logger.setLevel(logging.ERROR)
        no_timestamp_stream_logger.is_silent = True

    if log_dir:
        make_sure_path_exists(log_dir)
        timestamp_file_logger = logging.FileHandler(
            os.path.join(log_dir, log_file), 'a')
        timestamp_file_logger.setFormatter(log_format)
        timestamp_logger.addHandler(timestamp_file_logger)

        no_timestamp_file_logger = logging.FileHandler(
            os.path.join(log_dir, log_file), 'a')
        no_timestamp_file_logger.setFormatter(None)
        no_timestamp_logger.addHandler(no_timestamp_file_logger)

    timestamp_logger.info(f'{program_name} v{version}')
    timestamp_logger.info(ntpath.basename(
        sys.argv[0]) + ' ' + ' '.join(sys.argv[1:]))


def check_file_exists(input_file):
    """Check if file exists."""

    if not os.path.exists(input_file) or not os.path.isfile(input_file):
        logger = logging.getLogger('timestamp')
        logger.error('Input file does not exists: ' + input_file + '\n')
        sys.exit(1)


def check_dir_exists(input_dir):
    """Check if directory exists."""

    if not os.path.exists(input_dir) or not os.path.isdir(input_dir):
        logger = logging.getLogger('timestamp')
        logger.error('Input directory does not exists: ' + input_dir + '\n')
        sys.exit(1)


def make_sure_path_exists(path):
    """Create directory if it does not exist."""

    if not path:
        # lack of a path qualifier is acceptable as this
        # simply specifies the current directory
        return

    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            logger = logging.getLogger('timestamp')
            logger.error('Specified path could not be created: ' + path + '\n')
            sys.exit(1)


def is_executable(fpath):
    """Check if file is executable.
    This is a Python implementation of the linux
    command 'which'.

    Parameters
    ----------
    fpath : str
        Path to file.

    Returns
    -------
    boolean
        True if executable, else False.
    """

    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    """Return path to program.

    This is a Python implementation of the linux
    command 'which'.
    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    Parameters
    ----------
    program : str
        Name of executable for program.

    Returns
    -------
    str
        Path to executable, or None if it isn't on the path.
    """

    fpath, _fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file

    return None


def check_on_path(program, exit_on_fail=True):
    """Check if program is on the system path.

    Parameters
    ----------
    program : str
        Name of executable for program.
    exit_on_fail : boolean
        Exit program with error code -1 if program in not on path.

    Returns
    -------
    boolean
        True if program is on path, else False.
    """

    if which(program):
        return True

    if exit_on_fail:
        print('%s is not on the system path.' % program)
        sys.exit(1)

    return False


def check_dependencies(programs, exit_on_fail=True):
    """Check if all required programs are on the system path.

    Parameters
    ----------
    programs : iterable
        Names of executable programs.
    exit_on_fail : boolean
        Exit program with error code -1 if any program in not on path.

    Returns
    -------
    boolean
        True if all programs are on path, else False.
    """

    for program in programs:
        if not check_on_path(program, exit_on_fail):
            return False

    return True


def alphanumeric_sort(l):
    """Sorts the given iterable alphanumerically.

    http://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python

    Parameters
    ----------
    l : iterable
        The iterable to be sorted.

    Returns
    -------
    iterable
        Iterable sorted alphanumerically.
    """
    def convert(text): return int(text) if text.isdigit() else text

    def alphanum_key(key): return [convert(c)
                                   for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)



#!/usr/bin/env python3

from setuptools import setup, find_packages

import os


def version():
    setup_dir = os.path.dirname(os.path.realpath(__file__))
    version_file = open(os.path.join(setup_dir, 'unitem', 'VERSION'))
    return version_file.readline().strip()


if __name__ == '__main__':

    dirName = os.path.dirname(__file__)
    if dirName and os.getcwd() != dirName:
        os.chdir(dirName)

    setup(
        name='unitem',
        version=version(),
        author='Donovan Parks',
        author_email='donovan.parks@gmail.com',
        maintainer='Donovan Parks',
        maintainer_email='donovan.parks@gmail.com',
        packages=find_packages(),
        scripts=['bin/unitem'],
        package_data={'unitem': ['VERSION', '../markers/*']},
        url='http://pypi.python.org/pypi/unitem/',
        license='GPLv3',
        description='Ensemble binning strategies for combining the output of multiple binning methods.',
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Intended Audience :: Science/Research',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires=["numpy>=1.0", "svgwrite>=1.1.9"],
    )



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from .HMMMatch import HMMMatch


class HMMUnit(HMMMatch):
    """
    This class has been adapted from the Perl module written by Genome Research Ltd.
    Perl authors: Rob Finn (rdf@sanger.ac.uk), John Tate (jt6@sanger.ac.uk)
    Perl version: ?
    Python authors: Aaron Mussig (a.mussig@uq.edu.au)
    """

    def __init__(self):
        HMMMatch.__init__(self)
        self.proteinCoos = None
        self.seqEvalue = None
        self.domain = None
        self.seqFrom = None
        self.seqTo = None
        self.domEvalue = None
        self.hmmalign = dict()
        self.hmmFrom = None
        self.hmmTo = None
        self.envFrom = None
        self.envTo = None
        self.coreFrom = None
        self.coreTo = None
        self.aliAcc = None
        self.sig = None



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


class HMMMatch(object):
    """
    This class has been adapted from the Perl module written by Genome Research Ltd.
    Perl authors: Jaina Mistry (jm14@sanger.ac.uk), John Tate (jt6@sanger.ac.uk), Rob Finn (finnr@janelia.hhmi.org)
    Perl version: ?
    Python authors: Aaron Mussig (a.mussig@uq.edu.au)
    """

    def __init__(self):
        """
        evalue (float), bits (float), name (str) , bias (float)
        """
        self.evalue = None
        self.bits = None
        self.name = None
        self.bias = None



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import copy
import random
import gzip
from shutil import copyfile
from collections import defaultdict

from unitem.defaults import *
from unitem.common import (read_bins,
                           calculateN50L50M50)
from unitem.utils import make_sure_path_exists
from unitem.markers import Markers
from unitem.plot_common_bases import PlotCommonBases
from unitem.tree_common_bases import TreeCommonBases


class Ensemble():
    """Perform ensemble binning of genomes across multiple binning methods."""

    def __init__(self, bin_prefix):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

        self.bin_prefix = bin_prefix

    def _update_bins(self, bins, cids_to_remove):
        """Remove specified contigs from all bins.

        Parameters
        ----------
        bins : d[binning method][bin ID]
          Contigs for bins across all binning methods.
        cids_to_remove : iterable
          Contigs to remove from bins.
        """

        cids_to_remove = set(cids_to_remove)

        # remove scaffolds in highest quality bin from all other bins
        for method_id in bins:
            for bin_id in bins[method_id]:
                bins[method_id][bin_id] -= cids_to_remove

    def _update_gene_tables(self, gene_tables, cids_to_remove):
        """Remove specified contigs from marker gene tables.

        Parameters
        ----------
        gene_tables : d[binning method] -> gene table
          Marker gene tables for bins in each binning method.
        cids_to_remove : iterable
          Contigs to remove from marker gene tables.
        """

        cids_to_remove = set(cids_to_remove)

        # remove contigs from other bins
        for bm_gene_tables in gene_tables.values():
            for bin_id in bm_gene_tables:
                updated_table = defaultdict(list)
                for marker_id, contig_ids in bm_gene_tables[bin_id].items():
                    for cid in contig_ids:
                        if cid not in cids_to_remove:
                            updated_table[marker_id].append(cid)

                bm_gene_tables[bin_id] = updated_table

    def _bin_quality(self, bins, contigs, gene_tables, quality_weight, markers):
        """Determine estimated completeness, contamination, and quality of bins.

        Parameters
        ----------
        bins : d[binning method][bin ID] -> set(cid1, cid2, ... cidN)
          Contigs for bins across all binning methods.
        contigs : d[cid] -> seq
          Contigs across all bins.
        gene_tables : d[binning method] -> gene table
          Marker gene tables for bins in each binning method.
        quality_weight : float
          Weight given to contamination when assessing genome quality.
        markers : Markers
          Marker genes used to evaluate quality of genomes.

        Return
        ------
          List with bin metadata sorted by quality, then N50, the genome size.
        """

        q = []
        for binning_method in gene_tables:
            for bid, gene_table in gene_tables[binning_method].items():
                domain, comp, cont = markers.evaluate(gene_table)

                bin_seqs = [contigs[cid] for cid in bins[binning_method][bid]]
                n50, _l50, _m50 = calculateN50L50M50(bin_seqs)
                genome_size = sum([len(s) for s in bin_seqs])

                q.append((binning_method,
                          bid,
                          domain,
                          comp,
                          cont,
                          comp - quality_weight*cont,
                          n50,
                          genome_size))

        # sort bins by quality follwed by N50 followed by genome size, and
        # break remaining ties randomly
        q.sort(key=lambda x: (x[5], x[6], x[7], random.random()),
               reverse=True)

        return q

    def _matched_bin_sets(self, bins, contig_lens, bin_quality, min_quality, min_comp, max_cont, no_bin_matching):
        """Determine all sets of matched bins.

        Parameters
        ----------
        bins : d[binning method][bin ID]
          Contigs for bins across all binning methods.
        contig_lens : d[cid] -> length of contig
          Length of contigs.
        bin_quality : list of tuples
          Bin quality information. Must be sorted by quality!
        min_quality : float
          Minimum quality of bin to consider during bin matching.
        no_bin_matching : boolean
          Flag indicating bin matching should be skipped and all bins treated independently.

        Return
        ------
          Matched bin sets.
        """

        matched_sets = []
        processed_bins = defaultdict(set)
        for cur_bm, cur_bid, _domain, comp, cont, quality, N50, gs in bin_quality:
            if cur_bid in processed_bins[cur_bm]:
                continue  # bin has already been considered

            matched_bins = []
            matched_bins.append((cur_bm, cur_bid, quality, N50, gs))
            if not no_bin_matching and (quality >= min_quality and comp >= min_comp and cont <= max_cont):
                for test_bm, test_bid, _domain, comp, cont, quality, N50, gs in bin_quality:
                    if test_bm == cur_bm:
                        # can't group with a bin from the same binning method
                        continue

                    if test_bid in processed_bins[test_bm]:
                        continue  # bin has already been considered

                    if quality < min_quality or comp < min_comp or cont > max_cont:
                        continue

                    bp_in_common, total_bp1, total_bp2 = self._bases_in_common(bins[cur_bm][cur_bid],
                                                                               bins[test_bm][test_bid],
                                                                               contig_lens)

                    per_bp_in_common = bp_in_common / max(total_bp1, total_bp2)
                    if per_bp_in_common > 0.5:
                        matched_bins.append(
                            (test_bm, test_bid, quality, N50, gs))

            # removed matched bins
            for bm, bid, _q, _n50, _gs in matched_bins:
                processed_bins[bm].add(bid)

            matched_sets.append(tuple(matched_bins))

        # sort by total quality of bins in matched sets
        matched_sets.sort(key=lambda ms: (sum([x[2] for x in ms]),
                                          sum([x[3] for x in ms]),
                                          sum([x[4] for x in ms]),
                                          random.random()),
                          reverse=True)

        return matched_sets

    def _bases_in_common(self, contig_ids1, contig_ids2, contig_lens):
        """Calculate number of base pairs in common between two bins.

        Parameters
        ----------
        contig_ids1 : iterable
          Contigs in bin.
        contig_ids2 : iterable
          Contigs in bin.
        contig_lens : d[seq ID] -> length
          Length of contigs.

        Returns
        -------
          Base pairs in common, total bases in bin 1, total bases in bin 2
        """

        bp_in_common = sum([contig_lens[seq_id]
                            for seq_id in contig_ids2
                            if seq_id in contig_ids1])

        total_bp1 = sum([contig_lens[seq_id] for seq_id in contig_ids1])
        total_bp2 = sum([contig_lens[seq_id] for seq_id in contig_ids2])

        return bp_in_common, total_bp1, total_bp2

    def _resolve_matched_set(self,
                             matched_set,
                             bins,
                             contigs,
                             bin_quality,
                             remove_perc,
                             add_perc,
                             add_matches,
                             sel_min_quality,
                             sel_min_comp,
                             sel_max_cont,
                             greedy,
                             unanimous):
        """Select contigs to cluster from matched bin set.

        Parameters
        ----------
        matched_set : iterable with bin metadata (binning method, bin ID, quality)
          Matched set of bins. Highest quality bin must be first.
        bins : d[binning method][bin ID]
          Contigs for bins across all binning methods.
        contigs : d[cid] -> seq
          Contigs across all bins.
        bin_quality : list of tuples
          Bin quality information. Must be sorted by quality!
        remove_perc : float
          Minimum percentage of bins from other binning methods require to remove contig in highest quality bin.
        add_perc : float
          Minimum percentage of matched bins required to add contig to highest quality bin.
        add_matches : float
          Minimum number of matched bins required to 'add' contigs.
        sel_min_quality : float
          Minimum quality of bins to consider for filtering contigs.
        greedy : boolean
          Perform greedy filtering.
        unanimous : boolean
          Perform unanimous filtering.
        """

        # identify contig count for bins in matched in set
        primary_bm, primary_bid, _q, _n50, _gs = matched_set[0]
        primary_contigs = bins[primary_bm][primary_bid]

        matched_contigs = defaultdict(int)
        matched_bin_ids = set()
        matched_methods = set()
        for bm, bid, _q, _n50, _gs in matched_set:
            matched_bin_ids.add(bm + bid)
            matched_methods.add(bm)
            for cid in bins[bm][bid]:
                matched_contigs[cid] += 1

        # add or remove contigs based on consensus filtering
        new_bin = {}
        removed_contigs = set()
        added_contigs = set()
        if unanimous:
            # remove all contigs from highest quality bin that are
            # not present in all other matched bins
            num_matched_bins = len(matched_set)
            for cid in primary_contigs:
                match_count = matched_contigs.get(cid, 0)
                if match_count == num_matched_bins:
                    new_bin[cid] = contigs[cid]
        elif greedy:
            for cid in primary_contigs:
                new_bin[cid] = contigs[cid]
        else:
            # identify contig count for bins not in matched set
            unmatched_contigs = defaultdict(int)
            binned_non_matched_method = defaultdict(int)
            for bm, bid, _domain, comp, cont, quality, _N50, _gs in bin_quality:
                if bm+bid in matched_bin_ids:
                    continue

                if quality < sel_min_quality or comp < sel_min_comp or cont > sel_max_cont:
                    continue

                for cid in bins[bm][bid]:
                    unmatched_contigs[cid] += 1

                    if bm not in matched_methods:
                        binned_non_matched_method[cid] += 1

            # remove contigs from highest quality bin iff it appears in a
            # consensus of non-matched bins
            # (highest quality bin with contig is also counted)
            num_binning_methods = len(bins)
            for cid in primary_contigs:
                unmatched_count = unmatched_contigs.get(cid, 0)

                per_other_bm = unmatched_count * 100 / num_binning_methods
                if per_other_bm <= remove_perc:
                    new_bin[cid] = contigs[cid]
                else:
                    removed_contigs.add(cid)

            # add new contigs to highest quality bin iff there is a suitable
            # number of matched bins and the contig is present in a consensus
            # of these matched bins
            num_matched_bins = len(matched_set)
            if num_matched_bins >= add_matches:
                for cid, match_count in matched_contigs.items():
                    if cid in primary_contigs:
                        continue

                    # number of bins in denominator is number of matched
                    # binning methods plus number of unmatched binning methods
                    # where contig is in a bin
                    bin_count = num_matched_bins + \
                        binned_non_matched_method.get(cid, 0)
                    p = match_count * 100 / bin_count
                    if p >= add_perc:
                        new_bin[cid] = contigs[cid]
                        added_contigs.add(cid)

        return new_bin, removed_contigs, added_contigs

    def _report_selection(self,
                          bin_name,
                          bin,
                          markers,
                          quality_weight,
                          binning_method,
                          original_bid,
                          num_matched_bins,
                          num_removed,
                          num_added,
                          greedy,
                          unanimous):
        """Create string with information about selected bin."""

        domain, comp, cont = markers.bin_quality(bin)
        quality = comp - quality_weight*cont

        genome_size = sum([len(seq) for seq in bin.values()])
        num_scaffolds = len(bin)
        n50, l50, m50 = calculateN50L50M50(bin.values())

        row = '%s\t%s\t%s\t%s\t%.2f\t%.2f\t%.2f' % (bin_name,
                                                    binning_method,
                                                    original_bid,
                                                    domain,
                                                    comp,
                                                    cont,
                                                    quality)

        row += '\t%d\t%d\t%d\t%d' % (genome_size,
                                     n50,
                                     l50,
                                     num_scaffolds)

        if not greedy:
            row += '\t%s\t%s' % (num_matched_bins,
                                 num_removed)
            if not unanimous:
                row += '\t%s' % num_added

        row += '\n'

        return row

    def _bin_summary(self,
                     sel_bins,
                     merged_bins,
                     quality_weight,
                     markers,
                     summary_file):
        """Summarize quality of reported bins."""

        bins = sel_bins.copy()
        bins.update(merged_bins)

        comp_cont_count = defaultdict(lambda: defaultdict(int))
        quality_count = defaultdict(int)
        total_comp = 0
        total_cont = 0
        total_q = 0
        for bid, bin in bins.items():
            domain, comp, cont = markers.bin_quality(bin)
            quality = comp - quality_weight*cont

            for test_comp in [90, 80, 70]:
                for test_cont in [5, 10]:
                    if comp >= test_comp and cont <= test_cont:
                        comp_cont_count[test_comp][test_cont] += 1

            for test_q in [90, 70, 50]:
                if quality >= test_q:
                    quality_count[test_q] += 1

            total_comp += comp
            total_cont += cont
            total_q += quality

        fout = open(summary_file, 'w')
        fout.write('Criteria\tNo. Bins\n')
        for comp in [90, 80, 70]:
            for cont in [5, 10]:
                fout.write('Comp. >= %d, cont. <= %d\t%d\n' %
                           (comp, cont, comp_cont_count[comp][cont]))

        for q in [90, 70, 50]:
            fout.write('Quality >= %d\t%d\n' % (q, quality_count[q]))

        fout.write('Total completeness\t%.1f\n' % total_comp)
        fout.write('Total contamination\t%.1f\n' % total_cont)
        fout.write('Total quality\t%.1f\n' % total_q)

        fout.close()

    def _perc_bases_in_common(self, query_bin, bins, contig_lens):
        """Determine percentage of base pairs in common between a bin and a set of bins.

        Parameters
        ----------
        query_bin : d[cid] -> contig
          Bin to match.
        bins : d[binning method][bin ID]
          Contigs for bins across all binning methods.
        contig_lens : d[cid] -> contig length
          Length of all contigs.

        Returns
        -------
        list of tuples
          Each list indicates the binning method, bin id, and percentage of common bases.

        Only cases where the percentage of common bases is >= 50% are returned.
        """

        matches = []
        for bm in bins:
            for bid in bins[bm]:
                bp_in_common, total_bp1, total_bp2 = self._bases_in_common(query_bin.keys(),
                                                                           bins[bm][bid],
                                                                           contig_lens)

                perc_common = bp_in_common * 100 / max(total_bp1, total_bp2)
                if perc_common >= 50:
                    matches.append((bm, bid, perc_common))

        matches.sort(key=lambda x: (x[2], random.random()), reverse=True)

        return matches

    def _read_init_bin_quality(self, profile_dir):
        """Read quality of each bin as determined by the profile command."""

        bq = defaultdict(lambda: {})
        for f in os.listdir(profile_dir):
            if not f.endswith('_quality.tsv'):
                continue

            method_id = f.replace('_quality.tsv', '')
            with open(os.path.join(profile_dir, f)) as f:
                f.readline()

                for line in f:
                    tokens = line.strip().split('\t')
                    bid, domain, comp, cont, _ = tokens
                    bq[method_id][bid] = (domain, float(comp), float(cont))

        return bq

    def _write_initial_contig_state(self,
                                    methods_sorted,
                                    contigs_in_bins,
                                    init_bin_quality,
                                    output_dir):
        """Record initial state of each contig."""

        fout = open(os.path.join(output_dir, 'contig_info_initial.tsv'), 'w')
        fout.write('Contig ID')
        for bm in methods_sorted:
            fout.write('\t' + bm)
        fout.write('\n')

        for cid in contigs_in_bins:
            fout.write(cid)
            for bm in methods_sorted:
                if bm not in contigs_in_bins[cid]:
                    fout.write('\tunbinned')
                else:
                    bid = contigs_in_bins[cid][bm]
                    domain, comp, cont = init_bin_quality[bm][bid]
                    fout.write('\t%s,%.1f,%.1f' % (bid, comp, cont))
            fout.write('\n')
        fout.close()

    def run(self,
            profile_dir,
            bin_dirs,
            marker_dir,
            quality_weight,
            sel_min_quality,
            sel_min_comp,
            sel_max_cont,
            remove_perc,
            add_perc,
            add_matches,
            greedy,
            unanimous,
            report_min_quality,
            simple_headers,
            output_dir):
        """Perform ensemble binning of genomes across multiple binning methods.

        Parameters
        ----------
        profile_dir : str
          Directory with bin profiles (output of 'profile' command).
        bin_dirs : list of str
            Directories containing bins from different binning methods.
        quality_weight : float
          Weight given to contamination when assessing quality of bins.
        sel_min_quality : float
          Minimum quality of bins to select.
        sel_min_comp : float
          Minimum completeness of bins to select.
        sel_max_cont : float
          Maximum contamination of bins to select.
        remove_perc : float
          Minimum percentage of bins from other binning methods require to remove contig in highest quality bin.
        add_perc : float
          Minimum percentage of matched bins required to add contig to highest quality bin.
        add_matches : float
          Minimum number of matched bins required to 'add' contigs.
        greedy : boolean
          Perform greedy bin selection.
        unanimous : boolean
          Perform unanimous bin selection.
        report_min_quality : float
          Minimum quality of bins to report.
        simple_headers : boolean
          Flag indicating that information should not be appended to the headers of bin FASTA files.
        output_dir : str
          Output directory.
        """

        markers = Markers(marker_dir)

        if greedy:
            self.logger.info("Greedy selection with quality_weight = {:.1f}, sel_min_quality = {:.1f}, sel_min_comp = {:.1f}, and sel_max_cont = {:.1f}.".format(
                quality_weight, sel_min_quality, sel_min_comp, sel_max_cont))
            remove_perc = 101
            add_perc = 101
        elif unanimous:
            self.logger.info("Unanimous selection with quality_weight = {:.1f}, sel_min_quality = {:.1f}, sel_min_comp = {:.1f}, and sel_max_cont = {:.1f}.".format(
                quality_weight, sel_min_quality, sel_min_comp, sel_max_cont))
        else:
            self.logger.info("Consensus selection with quality_weight ={:.1f}, sel_min_quality = {:.1f}, sel_min_comp = {:.1f}, and sel_max_cont = {:.1f}.".format(
                quality_weight, sel_min_quality, sel_min_comp, sel_max_cont))
            self.logger.info('Removing and adding contigs by consensus with remove_perc = {:.1f}, add_perc = {:.1f}, add_matches = {:,}.'.format(
                remove_perc, add_perc, add_matches))

        self.logger.info(
            'Reporting bins with a quality >= {:.1f}.'.format(report_min_quality))

        # get scaffold IDs in bins across all binning methods
        self.logger.info('Reading all bins.')
        bins, contigs, contigs_in_bins = read_bins(bin_dirs)
        methods_sorted = sorted(bins.keys())
        contig_lens = {cid: len(contigs[cid]) for cid in contigs}
        orig_bins = copy.deepcopy(bins)

        # get marker genes for bins across all binning methods
        self.logger.info('Identifying marker genes across all bins.')
        gene_tables = markers.marker_gene_tables(
            profile_dir,
            binning_methods=bins.keys()
        )

        # create output directories
        bin_dir = os.path.join(output_dir, 'bins')
        make_sure_path_exists(bin_dir)

        # get initial quality of bins
        self.logger.info('Reading initial quality of bins.')
        init_bin_quality = self._read_init_bin_quality(profile_dir)

        # write out initial state of each contig
        self.logger.info('Recording initial state of contigs.')
        self._write_initial_contig_state(methods_sorted,
                                         contigs_in_bins,
                                         init_bin_quality,
                                         output_dir)

        # perform consensus selection of bins
        header = 'UniteM Bin ID\tBinning Method\tBin ID'
        header += '\tMarker Domain\tCompleteness (%)\tContamination (%)\tQuality (%)'
        header += '\tGenome Size\tN50\tL50\tNo. Contigs'
        if not greedy:
            header += '\tNo. Matched Bins\tNo. Removed Contigs'
            if not unanimous:
                header += '\tNo. Added Contigs'
        header += '\n'

        fout = open(os.path.join(output_dir, 'bin_info.tsv'), 'w')
        fout.write(header)

        fout_matched = open(os.path.join(
            output_dir, 'matched_set_info.tsv'), 'w')
        if greedy:
            fout_matched.write(
                'No. Matched Bins\tBin ID\tCompleteness (%)\tContamination (%)\tQuality (%)\n')
        else:
            fout_matched.write(
                'No. Matched Bins\tUniteM Bin ID\tCompleteness (%)\tContamination (%)\tQuality (%)')
            fout_matched.write(
                '\tMatched Bin ID\tCompleteness (%)\tContamination (%)\tQuality (%)\n')

        fout_bin_info = open(os.path.join(
            output_dir, 'matched_bin_info.tsv'), 'w')
        fout_bin_info.write(
            'UniteM Bin ID\tCompleteness (%)\tContamination (%)\tQuality (%)')
        fout_bin_info.write(
            '\tNo. Matched Bins\tMatched Bin ID\tPercent Common Bases\tCompleteness (%)\tContamination (%)\tQuality (%)\n')

        fout_contigs = open(os.path.join(output_dir, 'contig_info.tsv'), 'w')
        fout_contigs.write(
            'UniteM Bin ID\tContig ID\tNo. Matched Bins\tNo. Unmatched Bins\tNo. Degenerate Bins\tNo. Unbinned')
        for method in methods_sorted:
            fout_contigs.write(f'\t{method}')
        fout_contigs.write('\n')

        plot_dir = os.path.join(output_dir, 'common_bases')
        make_sure_path_exists(plot_dir)

        # perform ensemble binning
        self.logger.info('Performing ensemble binning.')
        bin_num = 0
        total_comp = 0
        total_cont = 0
        total_quality = 0
        sel_bins = {}
        selected_rows = {}
        unitem_common_bases = defaultdict(lambda: defaultdict(int))
        unitem_bin_quality = {}
        sanity_check_contigs = set()
        tree_common_bases = TreeCommonBases()
        while True:
            # determine highest quality match bin set
            bin_quality = self._bin_quality(bins,
                                            contigs,
                                            gene_tables,
                                            quality_weight,
                                            markers)

            matched_sets = self._matched_bin_sets(bins,
                                                  contig_lens,
                                                  bin_quality,
                                                  sel_min_quality,
                                                  sel_min_comp,
                                                  sel_max_cont,
                                                  greedy)

            if len(matched_sets) == 0:
                break  # no bins meeting selection criteria

            new_bin, removed_contigs, added_contigs = self._resolve_matched_set(
                matched_sets[0],
                bins,
                contigs,
                bin_quality,
                remove_perc,
                add_perc,
                add_matches,
                sel_min_quality,
                sel_min_comp,
                sel_max_cont,
                greedy,
                unanimous)

            _domain, comp, cont = markers.bin_quality(new_bin)

            quality = comp - quality_weight*cont
            if quality < report_min_quality:
                break

            total_comp += comp
            total_cont += cont
            total_quality += quality

            # report selection
            bin_num += 1
            unitem_bin_id = f'{self.bin_prefix}{bin_num}'
            unitem_bin_quality[unitem_bin_id] = (comp, cont)
            primary_bm, primary_bid, _q, _n50, _gs = matched_sets[0][0]
            self.logger.info("Selected {} from {} with quality = {:.1f} (comp. = {:.1f}%, cont. = {:.1f}%).".format(
                primary_bid,
                primary_bm,
                quality,
                comp,
                cont))

            if not greedy and not unanimous:
                # performing consensus binning
                self.logger.info("-> Identified {} matched bins, removed {} contigs, added {} contigs.".format(
                    len(matched_sets[0]),
                    len(removed_contigs),
                    len(added_contigs)))
            elif not greedy:
                # performing unanimous binning
                self.logger.info("-> Identified {} matched bins and removed {} contigs.".format(
                    len(matched_sets[0]),
                    len(removed_contigs)))

            # write out matched set
            fout_matched.write(f'{len(matched_sets[0])}')
            if not greedy:
                fout_matched.write('\t{}\t{:.1f}\t{:.1f}\t{:.1f}'.format(
                    unitem_bin_id,
                    comp,
                    cont,
                    quality))

            for bm, bid, q, _n50, _gs in matched_sets[0]:
                _domain, comp, cont = markers.bin_quality(bins[bm][bid])
                quality = comp - quality_weight*cont
                fout_matched.write('\t{}\t{:.1f}\t{:.1f}\t{:.1f}'.format(
                    bm + '~' + bid,
                    comp,
                    cont,
                    quality))
            fout_matched.write('\n')

            # write out common base pairs
            matches = self._perc_bases_in_common(
                new_bin, orig_bins, contig_lens)
            fout_bin_info.write('{}\t{:.1f}\t{:.1f}\t{:.1f}\t{}'.format(
                unitem_bin_id,
                comp,
                cont,
                quality,
                len(matches)))

            for bm, bid, perc_common in matches:
                unitem_common_bases[unitem_bin_id][bm] = perc_common
                _domain, comp, cont = markers.bin_quality(orig_bins[bm][bid])
                quality = comp - quality_weight*cont
                fout_bin_info.write('\t{}\t{:.1f}\t{:.1f}\t{:.1f}\t{:.1f}'.format(
                    bm + '~' + bid,
                    perc_common,
                    comp,
                    cont,
                    quality))

            fout_bin_info.write('\n')

            output_file = os.path.join(plot_dir, unitem_bin_id + '.svg')
            tree_common_bases.plot(unitem_bin_id, matches, output_file)

            # write out summary information about selected bins
            row = self._report_selection(unitem_bin_id,
                                         new_bin,
                                         markers,
                                         quality_weight,
                                         primary_bm,
                                         primary_bid,
                                         str(len(matched_sets[0])),
                                         str(len(removed_contigs)),
                                         str(len(added_contigs)),
                                         greedy,
                                         unanimous)
            fout.write(row)
            selected_rows[unitem_bin_id] = row

            # write out contig info
            matched_bins = {}
            for m in matched_sets[0]:
                matched_bins[m[0]] = m[1]

            new_bin_file = os.path.join(bin_dir, unitem_bin_id + '.fna.gz')
            fout_bin = gzip.open(new_bin_file, 'wt')
            for cid, seq in new_bin.items():
                fout_contigs.write(f'{unitem_bin_id}\t{cid}')

                if cid in sanity_check_contigs:
                    self.logger.error(f'Contig selected twice: {cid}')
                    sys.exit(1)
                sanity_check_contigs.add(cid)

                matched = 0
                unmatched = 0
                unbinned = 0
                degenerate = 0
                row = ''
                for m in methods_sorted:
                    if m in contigs_in_bins[cid]:
                        bid = contigs_in_bins[cid][m]
                        domain, comp, cont = markers.bin_quality(bins[m][bid])
                        if m in matched_bins and bid == matched_bins[m]:
                            row += '\t{},{},{:.1f},{:.1f}'.format(
                                'matched', bid, comp, cont)
                            matched += 1
                        else:
                            q = comp - quality_weight*cont
                            if q < sel_min_quality or comp < sel_min_comp or cont > sel_max_cont:
                                row += '\t{},{},{:.1f},{:.1f}'.format(
                                    'degenerate', bid, comp, cont)
                                degenerate += 1
                            else:
                                row += '\t{},{},{:.1f},{:.1f}'.format(
                                    'unmatched', bid, comp, cont)
                                unmatched += 1
                    else:
                        row += '\tunbinned'
                        unbinned += 1

                fout_contigs.write('\t{}\t{}\t{}\t{}'.format(
                    matched,
                    unmatched,
                    degenerate,
                    unbinned))
                fout_contigs.write(row + '\n')

                if simple_headers:
                    fout_bin.write(f'>{cid}\n')
                else:
                    cid_info = '[matched={}] [unmatched={}] [degenerate={}] [unbinned={}]'.format(
                        matched,
                        unmatched,
                        degenerate,
                        unbinned)

                    if cid in added_contigs:
                        cid_info += ' [added by consensus]'

                    fout_bin.write(f'>{cid} {cid_info}\n')
                fout_bin.write(seq + '\n')
            fout_bin.close()

            # remove contigs in highest quality bin from marker gene tables and all other bins
            self._update_gene_tables(gene_tables, new_bin.keys())
            self._update_bins(bins, new_bin.keys())

            sel_bins[unitem_bin_id] = new_bin

        self.logger.info(f'Selected {bin_num} bins.')
        self.logger.info('-> total comp. = {:.1f}, total cont. = {:.1f}, total quality = {:.1f}'.format(
            total_comp,
            total_cont,
            total_quality))

        fout.close()
        fout_matched.close()
        fout_contigs.close()
        fout_bin_info.close()

        plot = PlotCommonBases()
        output_plot = os.path.join(output_dir, 'percent_common_bases.svg')
        plot.plot(unitem_common_bases, unitem_bin_quality, output_plot)

        # summarize results of reported bins
        summary_file = os.path.join(output_dir, 'bin_quality_summary.tsv')
        self._bin_summary(sel_bins,
                          {},
                          quality_weight,
                          markers,
                          summary_file)
